'use strict';

/**
 * Lazy Properties
 */


/**
 * Shorthand helper for building descriptors up from a prototype chain.
 * @param {*} proto
 * @param {...any} objs
 * @returns {Object}
 */
function createAssign(proto, ...objs) {
  return Object.assign(Object.create(proto), ...objs);
}

/**
 * Defaults common to both data and accessor descriptors.
 * Null prototype.
 */
const defaultCommonDescriptor = createAssign(null, {
  configurable: true,
  enumerable: true,
});

/**
 * Defaults for an eventually-initialized property.
 * Same as if a property was set normally.
 */
const defaultDataDescriptor = createAssign(defaultCommonDescriptor, {
  writable: true,
  value: undefined,
});

/**
 * Defaults for an interim accessor property.
 */
const defaultAccessorDescriptor = createAssign(defaultCommonDescriptor, {
  get: undefined,
  set: undefined,
});

/**
 * Defer calling initializer to set value for name until name is read.
 * If a lazy property is defined on an object which is used as a prototype,
 * the objectBound flag determines whether the lazy initializer will be
 * invoked just once and set the property value on the original prototype
 * object, or for any inherited object and set the property on that object.
 * The objectBound flag also controls the 'this' object of the initializer
 * when called.
 * @param {Object} obj
 * @param {String} name
 * @param {() => {*}} initializer
 * @param {Object=} descriptor
 * @param {Boolean=} descriptor.configurable
 * @param {Boolean=} descriptor.enumerable
 * @param {Boolean=} descriptor.writable
 * @param {Boolean=} objectBound
 */
function lazy(obj, name, initializer, descriptor, objectBound = true) {
  if (typeof initializer !== 'function') {
    throw new TypeError('initializer is not callable');
  }

  // The descriptor for the eventually-initialized property.
  const finalDescriptor = createAssign(defaultDataDescriptor, descriptor);

  // The descriptor defining the interim accessor property.
  const lazyDescriptor = createAssign(defaultAccessorDescriptor, {
    // The accessor replaces itself with the value returned by invoking the initializer.
    get: function () {
      const targetObject = objectBound ? obj : this;
      finalDescriptor.value = initializer.apply(targetObject);
      Object.defineProperty(targetObject, name, finalDescriptor);
      return finalDescriptor.value;
    },
  
    /**
     * Unless a non-writable descriptor was specified, an explicit set will overwrite
     * the lazy descriptor with a default data property,
     */
    ...(finalDescriptor.writable && {
      set: function (value) {
        Object.defineProperty(this, name, createAssign(defaultDataDescriptor, {
          value,
        }))
      },
    }),
  
    configurable: true, // Ensure we can replace ourself.
    enumerable: finalDescriptor.enumerable, // Use requested visibility.
  });

  return Object.defineProperty(obj, name, lazyDescriptor);
}

module.exports = {
  lazy,
};
